/*!
 * node-restapi - version 0.0.8
 * Copyright(c) 2012 Ömer ÜCEL <omerucel@gmail.com>
 * MIT Licensed
 */

exports.version = '0.0.9';

var net         = require("net");
var events      = require("events");
var http        = require("http");
var HTTPParser  = process.binding("http_parser").HTTPParser;
var url         = require("url");
var qs          = require("qs");
var async       = require("async");

function RestError(code, data){
    if (code == undefined)
        code = 500;

    this.code = code;

    if (data == undefined)
        data = {};

    this.data = data;
}

RestError.prototype = new Error();
RestError.prototype.constructor = RestError;

/**
 * RouteClass
 */
var Route = function(){
    events.EventEmitter.call(this);
};

Route.super_ = events.EventEmitter;
Route.prototype = Object.create(events.EventEmitter.prototype, {
    constructor : {
        value : Route,
        enumerable : false
    }
});

Route.prototype.handle = function(stream){
    var app = new App();

    _events = this.listeners("preDispatch");
    for(i in _events)
    {
        app.on("preDispatch", _events[i]);
    }

    _events = this.listeners("postDispatch");
    for(i in _events)
    {
        app.on("postDispatch", _events[i]);
    }

    _events = this.listeners(stream._req.method);
    for(i in _events)
    {
        app.on(stream._req.method, _events[i]);
    }

    app.start(stream);
};

/**
 * AppClass
 */
var App = function(){
    events.EventEmitter.call(this);
};

App.super_ = events.EventEmitter;
App.prototype = Object.create(events.EventEmitter.prototype, {
    constructor : {
        value : App,
        enumerable : false
    }
});

App.prototype.stream = null;
App.prototype.start = function(stream){
    var self = this;
    self.stream = stream;
    self.req = stream._req;
    self.res = {};

    self.on("preDispatch", self.preDispatch);

    self.on("postDispatch", self.postDispatch);

    self.on("error", self.error);

    var tempList = [];

    for(i in (_list = self.listeners("preDispatch")))
        tempList.push(async.apply(_list[i], self));

    for(i in (_list = self.listeners(self.req.method)))
        tempList.push(async.apply(_list[i], self));

    for(i in (_list = self.listeners("postDispatch")))
        tempList.push(async.apply(_list[i], self));

    async.series(tempList, function(err, results){
        if (err)
            self.emit("error", err);
    });
};

App.prototype.setResponse = function(code, data){
    this.res = {
        code : code,
        data : data
    };
};

App.prototype.preDispatch = function(app, callback){
    callback(null);
};

App.prototype.postDispatch = function(app, callback){
    var bodyMessage = JSON.stringify(app.res);

    app.stream.write('HTTP/1.1 ' + app.res.code + ' ' + http.STATUS_CODES[app.res.code] + '\r\n');
    app.stream.write('Content-Length: ' + bodyMessage.length + '\r\n');
    app.stream.write('Content-Type: application/json; charset=utf-8\r\n');
    app.stream.write('\r\n' + bodyMessage + '\r\n');
    app.stream.end();
    callback(null);
};

App.prototype.error = function(err){
    this.setResponse(err.code, err.data);
    this.postDispatch(this, function(err){});
};

/**
 * RestApi Class
 */
var RestApi = function(){
    events.EventEmitter.call(this);
};

RestApi.super_ = events.EventEmitter;
RestApi.prototype = Object.create(events.EventEmitter.prototype, {
    constructor : {
        value : RestApi,
        enumerable : false
    }
});

RestApi.prototype.routes = {};

RestApi.prototype.createServer = function(){
    var self = this;

    var route404 = new Route();
    self.routes[404] = route404;
    route404.on("postDispatch", function(app, callback){
        app.setResponse(404, {});
        callback(null);
    });

    self.server = net.createServer(function(stream){
        stream.setTimeout(0);
        stream.setEncoding("utf8");

        stream.addListener("connect", function(){
            // İstek bilgileri tutulur.
            stream._req = {
                post : {},
                get : {},
                body : '',
                path : '',
                headers : {},
                method : 'GET',
                params : {}
            };

            // Http istekleri için parser hazırlanıyor.
            stream.parser = new HTTPParser(HTTPParser.REQUEST);

            // Gönderilen mesajı işler.
            stream.parser.onBody = function(buffer, start, len){
                stream._req.body += buffer.toString("utf8", start, start+len);
            };

            // Gönderilen datanın içindeki header bilgilerinin işlenmesi için bu metod çalışır.
            stream.parser.onHeadersComplete = function(info){
                var parsedUrl = url.parse(info.url);
                stream._req.method = info.method;
                stream._req.path = parsedUrl.pathname;
                stream._req.get = qs.parse(parsedUrl.query);
                stream._req.headers = info.headers;
            };

            // Gönderilen datanın işlenme işi bittiğinde bu metod çalışır.
            stream.parser.onMessageComplete = function(){
                stream._req.post = qs.parse(stream._req.body);
                self.handle(stream);
            };
        });

        stream.addListener("data", function(data){
            // Request parse ediliyor.
            var buffer = new Buffer(data);
            var ret = stream.parser.execute(buffer, 0, buffer.length);
            stream.parser.finish();
            if (ret instanceof Error)
                self.handle(stream);
        });
    });

    return self;
};

RestApi.prototype.listen = function(port){
    // TODO : Add host and listener.
    var self = this;
    self.server.listen(port);
};

RestApi.prototype.route = function(){
    var self = this;
    var routeClass = new Route();
    self.routes[arguments[0]] = routeClass;
    return routeClass;
};

RestApi.prototype.routes = function(routes){
    var self = this;
    for(route in routes)
        self.routes[route] = require(routes[route]).Route;
};

RestApi.prototype.handle = function(stream){
    var self        = this;
    var routeClass  = null;

    for(route in self.routes)
    {
        var re      = /(?:\?P\<([^>]*)\>)|(?::([a-zA-Z0-9_]+))/g;
        var params  = [];
        var _route  = route.replace(re, function(match) {
            params.push(arguments[1] || arguments[2]);
            return match.charAt(0) == ':' ? '([^/]+)' : '';
        });
        var match   = stream._req.path.match(_route);

        if (match)
        {
            for(var i = 0, param; (param = params[i]); i++)
                stream._req.params[param] = match[i+1];

            routeClass = self.routes[route];
            routeClass.handle(stream);
            break;
        }
    }

    if (routeClass == null)
    {
        routeClass = self.routes[404];
        routeClass.handle(stream);
    }
};

exports.createRoute = function(){
    return new Route();
};
exports.RestError = RestError;
exports.createServer = function(){
    var restApi = new RestApi();
    restApi.createServer();
    return restApi;
};